use super::{
    arith::ArithmeticOperation,
    cmp::CompOperation,
    modifiers::{MatchType, Mod},
    Operable,
};
use crate::{
    query::{IntoQuery, Metric},
    seal::Sealed,
};
use std::fmt::Display;

/// Represents a logical/set query operation (i.e. and, or, unless).
///
/// This type is meant to be only constructed through the provided methods on the
/// [``Metric``] type.
#[derive(Debug, Clone)]
pub struct LogicalOperation<'a, A: Operable, B: Operable> {
    op_type: LogicOp,
    left: A,
    right: B,
    matching: Mod<'a, MatchType>,
}

impl<'a, A: Operable, B: Operable> LogicalOperation<'a, A, B> {
    /// Match only given labels between the two vectors for evauluating the logical/set operation
    ///
    /// Note: Only to be used with vector expressions on both left and right
    /// i.e. `<vector expr> <bin-op> on(<label list>) <vector expr>`
    pub fn match_on(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.matching = Mod::from((MatchType::On, labels));
        self
    }

    /// Ignore given labels between the two vectors for evauluating the logical/set operation
    ///
    /// Note: Only to be used with vector expressions on both left and right
    /// i.e. `<vector expr> <bin-op> ignoring(<label list>) <vector expr>`
    pub fn ignoring(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.matching = Mod::from((MatchType::Ignoring, labels));
        self
    }
}

impl<A: Operable, B: Operable> Display for LogicalOperation<'_, A, B> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "({a} {op}{matching} {b})",
            a = self.left,
            b = self.right,
            op = self.op_type,
            matching = self.matching,
        )
    }
}

impl<A: Operable, B: Operable> Operable for LogicalOperation<'_, A, B> {}
impl<A: Operable, B: Operable> Sealed for LogicalOperation<'_, A, B> {}

impl<A: Operable, B: Operable> IntoQuery for LogicalOperation<'_, A, B> {
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.to_string()
    }
}

#[derive(Clone, Copy, Debug)]
enum LogicOp {
    And,
    Or,
    Unless,
}

impl Display for LogicOp {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            LogicOp::And => "and",
            LogicOp::Or => "or",
            LogicOp::Unless => "unless",
        })
    }
}

/// The trait that enables logical/set operations between two instant vectors
///
/// > Caution: Unlike other operator traits, this trait doesn't contain a blanket implementation for all types that implement [``Operable``].
/// > This is to prevent the [``crate::query::Scalar``] being used as an operand.
/// > Care must be taken when using these operators with complex queries that might evaluate to scalar values instead of instant vectors.
pub trait Logical<'a>: Operable
where
    Self: Sized,
{
    /// Apply the `and` logical binary operator to `Self` and other [``Operable``] type (note: Scalar type is not supported)
    ///
    /// The resulting query would look like `(self and right)`
    fn and<R: Operable>(self, right: R) -> LogicalOperation<'a, Self, R> {
        LogicalOperation {
            op_type: LogicOp::And,
            left: self,
            right,
            matching: Mod::default(),
        }
    }

    /// Apply the `or` logical binary operator to `Self` and other [``Operable``] type (note: Scalar type is not supported)
    ///
    /// The resulting query would look like `(self or right)`
    fn or<R: Operable>(self, right: R) -> LogicalOperation<'a, Self, R> {
        LogicalOperation {
            op_type: LogicOp::Or,
            left: self,
            right,
            matching: Mod::default(),
        }
    }

    /// Apply the `unless` logical binary operator to `Self` and other [``Operable``] type (note: Scalar type is not supported)
    ///
    /// The resulting query would look like `(self unless right)`
    fn unless<R: Operable>(self, right: R) -> LogicalOperation<'a, Self, R> {
        LogicalOperation {
            op_type: LogicOp::Unless,
            left: self,
            right,
            matching: Mod::default(),
        }
    }
}

impl Logical<'_> for Metric<'_> {}
impl<A: Operable, B: Operable> Logical<'_> for LogicalOperation<'_, A, B> {}
impl<A: Operable, B: Operable> Logical<'_> for ArithmeticOperation<'_, A, B> {}
impl<A: Operable, B: Operable> Logical<'_> for CompOperation<'_, A, B> {}

#[cfg(test)]
mod tests {
    use crate::query::{ops::logic::Logical, Metric};

    #[test]
    fn logic_all_ops() {
        let query = Metric::new("one").and(Metric::new("two")).to_string();
        assert_eq!(query, "(one and two)");

        let query = Metric::new("one").or(Metric::new("two")).to_string();
        assert_eq!(query, "(one or two)");

        let query = Metric::new("one").unless(Metric::new("two")).to_string();
        assert_eq!(query, "(one unless two)");
    }

    #[test]
    fn logic_match_on() {
        let metric = Metric::new("test_metric")
            .and(Metric::new("another_metric"))
            .match_on(["one", "two"])
            .to_string();
        assert_eq!(metric, "(test_metric and on(one,two) another_metric)");
    }

    #[test]
    fn logic_match_ignoring() {
        let metric = Metric::new("test_metric")
            .and(Metric::new("another_metric"))
            .ignoring(["one", "two"])
            .to_string();
        assert_eq!(metric, "(test_metric and ignoring(one,two) another_metric)");
    }
}
